@@ -6,30 +6,41 @@ the properties and status of Zeek packages.
 import os
 import re
 
+from functools import total_ordering
+
+import semantic_version as semver
+
 from .uservar import UserVar
-from ._util import find_sentence_end
+from ._util import find_sentence_end, normalize_version_tag
 
 #: The name of files used by packages to store their metadata.
-METADATA_FILENAME = 'zkg.meta'
-LEGACY_METADATA_FILENAME = 'bro-pkg.meta'
+METADATA_FILENAME = "zkg.meta"
+LEGACY_METADATA_FILENAME = "bro-pkg.meta"
 
-TRACKING_METHOD_VERSION = 'version'
-TRACKING_METHOD_BRANCH = 'branch'
-TRACKING_METHOD_COMMIT = 'commit'
+TRACKING_METHOD_VERSION = "version"
+TRACKING_METHOD_BRANCH = "branch"
+TRACKING_METHOD_COMMIT = "commit"
+TRACKING_METHOD_BUILTIN = "builtin"
+
+BUILTIN_SOURCE = "zeek-builtin"
+BUILTIN_SCHEME = "zeek-builtin://"
+
+PLUGIN_MAGIC_FILE = "__zeek_plugin__"
+PLUGIN_MAGIC_FILE_DISABLED = "__zeek_plugin__.disabled"
+LEGACY_PLUGIN_MAGIC_FILE = "__bro_plugin__"
+LEGACY_PLUGIN_MAGIC_FILE_DISABLED = "__bro_plugin__.disabled"
 
-PLUGIN_MAGIC_FILE = '__bro_plugin__'
-PLUGIN_MAGIC_FILE_DISABLED = '__bro_plugin__.disabled'
 
 def name_from_path(path):
     """Returns the name of a package given a path to its git repository."""
-    return canonical_url(path).split('/')[-1]
+    return canonical_url(path).split("/")[-1]
 
 
 def canonical_url(path):
     """Returns the url of a package given a path to its git repo."""
-    url = path.rstrip('/')
+    url = path.rstrip("/")
 
-    if url.startswith('.') or url.startswith('/'):
+    if url.startswith(".") or url.startswith("/"):
         url = os.path.realpath(url)
 
     return url
@@ -41,6 +52,14 @@ def is_valid_name(name):
         # Reject names with leading/trailing whitespace
         return False
 
+    # For aliases: Do not allow file separators.
+    if "/" in name:
+        return False
+
+    # Avoid creating hidden files and directories.
+    if name.startswith("."):
+        return False
+
     if name in ("package", "packages"):
         return False
 
@@ -49,38 +68,38 @@ def is_valid_name(name):
 
 def aliases(metadata_dict):
     """Return a list of package aliases found in metadata's 'aliases' field."""
-    if 'aliases' not in metadata_dict:
+    if "aliases" not in metadata_dict:
         return []
 
-    return re.split(',\s*|\s+', metadata_dict['aliases'])
+    return re.split(r",\s*|\s+", metadata_dict["aliases"])
 
 
 def tags(metadata_dict):
     """Return a list of tag strings found in the metadata's 'tags' field."""
-    if 'tags' not in metadata_dict:
+    if "tags" not in metadata_dict:
         return []
 
-    return re.split(',\s*', metadata_dict['tags'])
+    return re.split(r",\s*", metadata_dict["tags"])
 
 
 def short_description(metadata_dict):
     """Returns the first sentence of the metadata's 'desciption' field."""
-    if 'description' not in metadata_dict:
-        return ''
+    if "description" not in metadata_dict:
+        return ""
 
-    description = metadata_dict['description']
-    lines = description.split('\n')
-    rval = ''
+    description = metadata_dict["description"]
+    lines = description.split("\n")
+    rval = ""
 
     for line in lines:
         line = line.lstrip()
-        rval += ' '
+        rval += " "
         period_idx = find_sentence_end(line)
 
         if period_idx == -1:
             rval += line
         else:
-            rval += line[:period_idx + 1]
+            rval += line[: period_idx + 1]
             break
 
     return rval.lstrip()
@@ -103,7 +122,7 @@ def user_vars(metadata_dict):
     return [(uvar.name(), uvar.val(), uvar.desc()) for uvar in uvars]
 
 
-def dependencies(metadata_dict, field='depends'):
+def dependencies(metadata_dict, field="depends"):
     """Returns a dictionary of (str, str) based on metadata's dependency field.
 
     The keys indicate the name of a package (shorthand name or full git URL).
@@ -134,7 +153,65 @@ def dependencies(metadata_dict, field='depends'):
     return rval
 
 
-class InstalledPackage(object):
+class PackageVersion:
+    """
+    Helper class to compare package versions with version specs.
+    """
+
+    def __init__(self, method, version):
+        self.method = method
+        self.version = version
+        self.req_semver = None
+
+    def fullfills(self, version_spec):
+        """
+        Whether this package version fullfills the given version_spec.
+
+        Returns:
+            (some message, bool)
+        """
+        if version_spec == "*":  # anything goes
+            return "", True
+
+        if self.method == TRACKING_METHOD_COMMIT:
+            return f'tracking method commit not compatible with "{version_spec}"', False
+
+        elif self.method == TRACKING_METHOD_BRANCH:
+            return "tracking method branch and commit", False
+            if version_spec.startswith("branch="):
+                branch = version_spec[len("branch=") :]
+
+            if version_spec != self.version:
+                return f'branch "{self.version}" not matching "{branch}"', False
+
+            return "", True
+
+        else:
+            # TRACKING_METHOD_BRANCH / TRACKING_METHOD_BUILTIN
+            if version_spec.startswith("branch="):
+                branch = version_spec[len("branch=") :]
+                return (
+                    f"branch {branch} requested, but using method {self.method}",
+                    False,
+                )
+
+            if self.req_semver is None:
+                normal_version = normalize_version_tag(self.version)
+                self.req_semver = semver.Version.coerce(normal_version)
+
+            try:
+                semver_spec = semver.Spec(version_spec)
+            except ValueError:
+                return f'invalid semver spec: "{version_spec}"', False
+            else:
+                if self.req_semver in semver_spec:
+                    return "", True
+
+                return f"{self.version} not in {version_spec}", False
+
+
+@total_ordering
+class InstalledPackage:
     """An installed package and its current status.
 
     Attributes:
@@ -147,11 +224,31 @@ class InstalledPackage(object):
         self.package = package
         self.status = status
 
+    def __repr__(self):
+        return f"InstalledPackage(package={self.package!r}, status={self.status!r})"
+
+    def __hash__(self):
+        return hash(str(self.package))
+
+    def __eq__(self, other):
+        return str(self.package) == str(other.package)
+
     def __lt__(self, other):
         return str(self.package) < str(other.package)
 
+    def is_builtin(self):
+        return self.package.is_builtin()
 
-class PackageStatus(object):
+    def fullfills(self, version_spec):
+        """
+        Does the current version fullfill version_spec?
+        """
+        return PackageVersion(
+            self.status.tracking_method, self.status.current_version
+        ).fullfills(version_spec)
+
+
+class PackageStatus:
     """The status of an installed package.
 
     This class contains properties of a package related to how the package
@@ -164,10 +261,11 @@ class PackageStatus(object):
 
         is_outdated (bool): whether a newer version of the package exists.
 
-        tracking_method (str): either "branch", "version", or "commit" to
-            indicate (respectively) whether package upgrades should stick to a
-            git branch, use git version tags, or do nothing because the
-            package is to always use a specific git commit hash.
+        tracking_method (str): either "branch", "version", "commit", or
+            "builtin" to indicate (respectively) whether package upgrades
+            should stick to a git branch, use git version tags, do nothing
+            because the package is to always use a specific git commit hash,
+            or do nothing because the package is built into Zeek.
 
         current_version (str): the current version of the installed
             package, which is either a git branch name or a git version tag.
@@ -176,8 +274,15 @@ class PackageStatus(object):
             package's current version/commit.
     """
 
-    def __init__(self, is_loaded=False, is_pinned=False, is_outdated=False,
-                 tracking_method=None, current_version=None, current_hash=None):
+    def __init__(
+        self,
+        is_loaded=False,
+        is_pinned=False,
+        is_outdated=False,
+        tracking_method=None,
+        current_version=None,
+        current_hash=None,
+    ):
         self.is_loaded = is_loaded
         self.is_pinned = is_pinned
         self.is_outdated = is_outdated
@@ -185,8 +290,12 @@ class PackageStatus(object):
         self.current_version = current_version
         self.current_hash = current_hash
 
+    def __repr__(self):
+        member_str = ", ".join(f"{k}={v!r}" for (k, v) in self.__dict__.items())
+        return f"PackageStatus({member_str})"
 
-class PackageInfo(object):
+
+class PackageInfo:
     """Contains information on an arbitrary package.
 
     If the package is installed, then its status is also available.
@@ -218,9 +327,18 @@ class PackageInfo(object):
             value is None.
     """
 
-    def __init__(self, package=None, status=None, metadata=None, versions=None,
-                 metadata_version='', invalid_reason='', version_type='',
-                 metadata_file=None, default_branch=None):
+    def __init__(
+        self,
+        package=None,
+        status=None,
+        metadata=None,
+        versions=None,
+        metadata_version="",
+        invalid_reason="",
+        version_type="",
+        metadata_file=None,
+        default_branch=None,
+    ):
         self.package = package
         self.status = status
         self.metadata = {} if metadata is None else metadata
@@ -250,7 +368,7 @@ class PackageInfo(object):
         This will be the first sentence of the package's 'description' field."""
         return short_description(self.metadata)
 
-    def dependencies(self, field='depends'):
+    def dependencies(self, field="depends"):
         """Returns a dictionary of dependency -> version strings.
 
         The keys indicate the name of a package (shorthand name or full git
@@ -286,8 +404,15 @@ class PackageInfo(object):
 
         return self.default_branch
 
+    def is_builtin(self):
+        return self.package and self.package.is_builtin()
 
-class Package(object):
+    def __repr__(self):
+        return f"PackageInfo(package={self.package!r}, status={self.status!r})"
+
+
+@total_ordering
+class Package:
     """A Zeek package.
 
     This class contains properties of a package that are defined by the package
@@ -318,8 +443,15 @@ class Package(object):
             be accurate/up-to-date).
     """
 
-    def __init__(self, git_url, source='', directory='', metadata=None,
-                 name=None, canonical=False):
+    def __init__(
+        self,
+        git_url,
+        source="",
+        directory="",
+        metadata=None,
+        name=None,
+        canonical=False,
+    ):
         self.git_url = git_url
         self.source = source
         self.directory = directory
@@ -340,7 +472,19 @@ class Package(object):
         return self.qualified_name()
 
     def __repr__(self):
-        return self.git_url
+        return (
+            f"Package(git_url={self.git_url!r}, source={self.source!r},"
+            f" directory={self.directory!r} name={self.name!r})"
+        )
+
+    def is_builtin(self):
+        return self.git_url.startswith(BUILTIN_SCHEME)
+
+    def __hash__(self):
+        return hash(str(self))
+
+    def __eq__(self, other):
+        return str(self) == str(other)
 
     def __lt__(self, other):
         return str(self) < str(other)
@@ -368,7 +512,7 @@ class Package(object):
         package has not been installed yet."""
         return short_description(self.metadata)
 
-    def dependencies(self, field='depends'):
+    def dependencies(self, field="depends"):
         """Returns a dictionary of dependency -> version strings.
 
         The keys indicate the name of a package (shorthand name or full git
@@ -402,7 +546,7 @@ class Package(object):
         just the package name is returned.
         """
         if self.directory:
-            return '{}/{}'.format(self.directory, self.name)
+            return f"{self.directory}/{self.name}"
 
         return self.name
 
@@ -414,8 +558,7 @@ class Package(object):
         git URL is returned.
         """
         if self.source:
-            return '{}/{}'.format(self.source,
-                                  self.name_with_source_directory())
+            return f"{self.source}/{self.name_with_source_directory()}"
 
         return self.git_url
 
@@ -425,11 +568,11 @@ class Package(object):
         E.g for a package with :meth:`qualified_name()` of "zeek/alice/foo",
         the following inputs will match: "foo", "alice/foo", "zeek/alice/foo"
         """
-        path_parts = path.split('/')
+        path_parts = path.split("/")
 
         if self.source:
             pkg_path = self.qualified_name()
-            pkg_path_parts = pkg_path.split('/')
+            pkg_path_parts = pkg_path.split("/")
 
             for i, part in reversed(list(enumerate(path_parts))):
                 ri = i - len(path_parts)
@@ -443,3 +586,32 @@ class Package(object):
                 return True
 
             return path == self.git_url
+
+
+def make_builtin_package(*, name: str, current_version: str, current_hash: str = None):
+    """
+    Given ``name``, ``version`` and ``commit`` as found in Zeek's ``zkg.provides``
+    entry, construct a :class:`PackageInfo` instance representing the built-in
+    package.
+    """
+    package = Package(
+        git_url=f"{BUILTIN_SCHEME}{name}",
+        name=name,
+        source=BUILTIN_SOURCE,
+        canonical=True,
+    )
+
+    status = PackageStatus(
+        is_loaded=True,  # May not hold in the future?
+        is_outdated=False,
+        is_pinned=True,
+        tracking_method=TRACKING_METHOD_BUILTIN,
+        current_version=current_version,
+        current_hash=current_hash,
+    )
+
+    return PackageInfo(
+        package=package,
+        status=status,
+        versions=[current_version],
+    )
